# pyright: reportAny=false
from __future__ import annotations

from typing import Any
from typing import cast

from django import template
from django.template.base import NodeList
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context
from django.template.loader import select_template
from django.utils.safestring import SafeString
from django.utils.safestring import mark_safe

from django_bird._typing import override
from django_bird.conf import app_settings

register = template.Library()


@register.tag(name="bird")
def do_bird(parser: Parser, token: Token) -> BirdNode:
    bits = token.split_contents()

    if len(bits) < 2:
        msg = f"{token.contents.split()[0]} tag requires at least one argument"
        raise template.TemplateSyntaxError(msg)

    # {% bird name %}
    # {% bird 'name' %}
    # {% bird "name" %}
    name = bits[1].strip("'\"")
    attrs = bits[2:]

    # self-closing tag
    # {% bird name / %}
    if len(attrs) > 0 and attrs[-1] == "/":
        nodelist = None
    else:
        nodelist = parser.parse(("endbird",))
        parser.delete_first_token()

    return BirdNode(name, attrs, nodelist)


class BirdNode(template.Node):
    def __init__(self, name: str, attrs: list[str], nodelist: NodeList | None) -> None:
        self.name = name
        self.attrs = attrs
        self.nodelist = nodelist
        self.default_slot = "default"

    @override
    def render(self, context: Context) -> SafeString:
        component_context = self.get_component_context_data(context)
        template_names = self.get_template_names()
        template = select_template(template_names)
        return template.render(component_context)

    def get_component_context_data(self, context: Context) -> dict[str, Any]:
        rendered_slots = self.render_slots(context)
        flat_attrs = self.flatten_attrs(context)
        default_slot = rendered_slots.get(self.default_slot) or context.get("slot")
        return {
            "attrs": mark_safe(flat_attrs),
            "slot": mark_safe(default_slot),
            "slots": {
                name: mark_safe(content) for name, content in rendered_slots.items()
            },
        }

    def render_slots(self, context: Context) -> dict[str, str]:
        if self.nodelist is None:
            return {}

        contents: dict[str, list[str]] = {self.default_slot: []}
        active_slot = self.default_slot

        for node in self.nodelist:
            if isinstance(node, SlotNode):
                active_slot = node.name
                contents.setdefault(active_slot, [])
            else:
                active_slot = self.default_slot

            rendered_content = node.render(context)
            contents[active_slot].append(rendered_content)

        if (
            all(not content for content in contents[self.default_slot])
            and "slot" in context
        ):
            contents[self.default_slot] = [context["slot"]]

        return {
            slot: template.Template("".join(content)).render(context)
            for slot, content in contents.items()
        }

    def flatten_attrs(self, context: Context) -> str:
        attrs: dict[str, str | None | bool] = {}

        for attr in self.attrs:
            if "=" in attr:
                key, value = attr.split("=", 1)
                attrs[key] = template.Variable(value).resolve(context)
            else:
                attrs[attr] = True

        return " ".join(
            key if isinstance(value, bool) and value else f'{key}="{value}"'
            for key, value in attrs.items()
        )

    def get_template_names(self):
        """
        Generate a list of potential template names for a component.

        The function searches for templates in the following order:
        1. In the base component directory, using the full component name
        2. In a subdirectory named after the component
            a. Using the full name (for subcomponents)
            b. Using the last part of the name
        3. In the same subdirectory, using a fallback 'index.html'

        The order of names is important as it determines the template resolution priority.
        This order allows for both direct matches and hierarchical component structures,
        with more specific paths taking precedence over more general ones.

        This order allows for:
        - Single file components
        - Multi-part components
        - Specific named files within component directories
        - Fallback default files for components

        For example:
        - For an "input" component, the ordering would be:
          1. `{component_dir}/input.html`
          2. `{component_dir}/input/input.html`
          3. `{component_dir}/input/index.html`
        - For an "input.label" component:
          1. `{component_dir}/input.label.html`
          2. `{component_dir}/input/label.html`
          3. `{component_dir}/input/label/label.html`
          3. `{component_dir}/input/label/index.html`

        Returns:
            list[str]: A list of potential template names in resolution order.
        """

        template_names: list[str] = []

        for component_dir in app_settings.COMPONENT_DIRS + ["bird"]:
            name_parts = self.name.split(".")
            path_name = "/".join(name_parts)
            potential_names = [
                f"{component_dir}/{self.name}.html",
                f"{component_dir}/{path_name}.html",
                f"{component_dir}/{path_name}/{name_parts[-1]}.html",
                f"{component_dir}/{path_name}/index.html",
            ]
            # Preserve order while ensuring uniqueness by removing duplicates.
            # Using a set here would not keep the ordering which is important
            # in template resolution.
            ordered_unique_names = dict.fromkeys(potential_names)
            template_names.extend(list(ordered_unique_names))

        return template_names


@register.tag("bird:slot")
def do_slot(parser: Parser, token: Token) -> SlotNode:
    bits = token.split_contents()
    name = parse_slot_name(bits)
    nodelist = parser.parse(("endbird:slot",))
    parser.delete_first_token()
    return SlotNode(name, nodelist)


def parse_slot_name(tag_args: list[str]) -> str:
    if len(tag_args) == 1:
        return "default"
    elif len(tag_args) == 2:
        name = tag_args[1]
        if name.startswith("name="):
            name = name.split("=")[1]
        else:
            name = name
        return name.strip("'\"")
    else:
        raise template.TemplateSyntaxError(
            "slot tag requires either no arguments, one argument, or 'name=\"slot_name\"'"
        )


class SlotNode(template.Node):
    def __init__(self, name: str, nodelist: NodeList):
        self.name = name
        self.nodelist = nodelist

    @override
    def render(self, context: Context) -> SafeString:
        default_content: str = self.nodelist.render(context)
        slots = context.get("slots")

        if not slots or not isinstance(slots, dict):
            slot_content = default_content
        else:
            slots_dict = cast(dict[str, str], slots)
            slot_content = slots_dict.get(self.name, default_content)

        # Recursively process the slot content
        t = template.Template(slot_content)
        content = t.render(context)

        return mark_safe(content)



from __future__ import annotations

import pytest
from django.template import Context
from django.template import Template
from django.template.base import Parser
from django.template.base import Token
from django.template.base import TokenType
from django.template.exceptions import TemplateSyntaxError
from django.test import override_settings

from django_bird.templatetags.django_bird import BirdNode
from django_bird.templatetags.django_bird import do_bird
from django_bird.templatetags.django_bird import parse_slot_name


class TestBirdTemplateTag:
    @pytest.mark.parametrize(
        "name_arg,expected",
        [
            ("button", "button"),
            ("'button'", "button"),
            ('"button"', "button"),
            ("button.label", "button.label"),
        ],
    )
    def test_node_name(self, name_arg, expected):
        token = Token(TokenType.BLOCK, f"bird {name_arg}")
        parser = Parser(
            [Token(TokenType.BLOCK, "endbird")],
        )
        node = do_bird(parser, token)
        assert node.name == expected

    def test_missing_argument(self):
        token = Token(TokenType.BLOCK, "bird")
        parser = Parser(
            [Token(TokenType.BLOCK, "endbird")],
        )
        with pytest.raises(TemplateSyntaxError):
            do_bird(parser, token)

    @pytest.mark.parametrize(
        "attrs,expected_attrs",
        [
            ("class='btn'", ["class='btn'"]),
            ("class='btn' id='my-btn'", ["class='btn'", "id='my-btn'"]),
            ("disabled", ["disabled"]),
        ],
    )
    def test_node_attrs(self, attrs, expected_attrs):
        token = Token(TokenType.BLOCK, f"bird button {attrs}")
        parser = Parser(
            [Token(TokenType.BLOCK, "endbird")],
        )
        node = do_bird(parser, token)
        assert node.attrs == expected_attrs

    @pytest.mark.parametrize(
        "component,template,expected",
        [
            (
                "<button>Click me</button>",
                "{% bird button %}Click me{% endbird %}",
                "<button>Click me</button>",
            ),
            (
                "<button>Click me</button>",
                "{% bird 'button' %}Click me{% endbird %}",
                "<button>Click me</button>",
            ),
            (
                "<button>Click me</button>",
                '{% bird "button" %}Click me{% endbird %}',
                "<button>Click me</button>",
            ),
        ],
    )
    def test_rendered_name(
        self, component, template, expected, create_bird_template, normalize_whitespace
    ):
        create_bird_template("button", component)
        t = Template(template)
        rendered = t.render(context=Context({}))
        assert normalize_whitespace(rendered) == expected

    @pytest.mark.parametrize(
        "component,template,context,expected",
        [
            (
                "<button {{ attrs }}>Click me</button>",
                "{% bird button class='btn' %}Click me{% endbird %}",
                {},
                '<button class="btn">Click me</button>',
            ),
            (
                "<button {{ attrs }}>Click me</button>",
                "{% bird button class='btn' id='my-btn' disabled %}Click me{% endbird %}",
                {},
                '<button class="btn" id="my-btn" disabled>Click me</button>',
            ),
            (
                "<button {{ attrs }}>{{ slot }}</button>",
                "{% bird button %}{{ slot }}{% endbird %}",
                {"slot": "Click me"},
                "<button>Click me</button>",
            ),
        ],
    )
    def test_rendered_attrs(
        self,
        component,
        template,
        context,
        expected,
        create_bird_template,
        normalize_whitespace,
    ):
        create_bird_template("button", component)
        t = Template(template)
        rendered = t.render(context=Context(context))
        assert normalize_whitespace(rendered) == expected

    def test_self_closing_tag(self, create_bird_template):
        create_bird_template(name="image", content="<img src='image' />")
        template = Template("{% bird image src='image' / %}")
        rendered = template.render(context=Context({}))
        assert rendered == "<img src='image' />"

    @pytest.mark.parametrize(
        "component,template,context,expected",
        [
            (
                "<button>{{ slot }}</button>",
                "{% bird button %}Click me{% endbird %}",
                {},
                "<button>Click me</button>",
            ),
            (
                "<button>{% bird:slot %}{% endbird:slot %}</button>",
                "{% bird button %}Click me{% endbird %}",
                {},
                "<button>Click me</button>",
            ),
            (
                "<button>{% bird:slot default %}{% endbird:slot %}</button>",
                "{% bird button %}Click me{% endbird %}",
                {},
                "<button>Click me</button>",
            ),
            (
                "<button><span>{% bird:slot leading-icon %}{% endbird:slot %}</span>{{ slot }}</button>",
                "{% bird button %}{% bird:slot leading-icon %}Icon here{% endbird:slot %}Click me{% endbird %}",
                {},
                "<button><span>Icon here</span>Click me</button>",
            ),
        ],
    )
    def test_with_slots(
        self,
        component,
        template,
        context,
        expected,
        create_bird_template,
        normalize_whitespace,
    ):
        create_bird_template("button", component)
        t = Template(template)
        rendered = t.render(context=Context(context))
        assert normalize_whitespace(rendered) == expected


@pytest.mark.xfail(reason="Feature not implemented yet")
class TestBirdTemplateTagFutureFeatures:
    def test_dynamic_attrs(self, create_bird_template):
        create_bird_template(
            name="button", content="<button class=btn_class>Click me</button>"
        )
        template = Template("{% bird button class=btn_class %}Click me{% endbird %}")
        rendered = template.render(context=Context({"btn_class": "primary"}))
        assert rendered == "<button class='primary'>Click me</button>"


class TestBirdNode:
    @pytest.mark.parametrize(
        "sub_dir,filename,component,nodename,expected",
        [
            (
                None,
                "button",
                "<button>Click me</button>",
                "button",
                "bird/button.html",
            ),
            (
                "button",
                "index",
                "<button>Click me</button>",
                "button",
                "bird/button/index.html",
            ),
            (
                "button",
                "button",
                "<button>Click me</button>",
                "button",
                "bird/button/button.html",
            ),
            (
                None,
                "button.label",
                "<button>Click me</button>",
                "button.label",
                "bird/button.label.html",
            ),
            (
                None,
                "button.label",
                "<button>Click me</button>",
                "button.label",
                "bird/button/label.html",
            ),
        ],
    )
    def test_get_template_names(
        self, sub_dir, filename, component, nodename, expected, create_bird_template
    ):
        create_bird_template(name=filename, content=component, sub_dir=sub_dir)
        node = BirdNode(name=nodename, attrs=[], nodelist=None)

        template_names = node.get_template_names()

        assert any(expected in template_name for template_name in template_names)

    @override_settings(DJANGO_BIRD={"COMPONENT_DIRS": ["not_default"]})
    def test_get_template_names_path_component_dirs(self, create_bird_template):
        create_bird_template(
            name="button",
            content="<button class=btn_class>Click me</button>",
            bird_dir_name="not_default",
        )

        node = BirdNode(name="button", attrs=[], nodelist=None)

        template_names = node.get_template_names()

        assert any("not_default" in template_name for template_name in template_names)


class TestSlotsTemplateTag:
    @pytest.mark.parametrize(
        "tag_args,expected",
        [
            (["slot"], "default"),
            (["slot", "foo"], "foo"),
            (["slot", "'foo'"], "foo"),
            (["slot", '"foo"'], "foo"),
            (["slot", 'name="foo"'], "foo"),
            (["slot", "name='foo'"], "foo"),
        ],
    )
    def test_parse_slot_name(self, tag_args, expected):
        assert parse_slot_name(tag_args) == expected

    def test_parse_slot_name_no_args(self):
        with pytest.raises(TemplateSyntaxError):
            assert parse_slot_name([])

    @pytest.mark.parametrize(
        "template,context,expected",
        [
            ("{{ slot }}", {"slot": "test"}, "test"),
            ("{% bird:slot %}{% endbird:slot %}", {"slot": "test"}, "test"),
            ("{% bird:slot default %}{% endbird:slot %}", {"slot": "test"}, "test"),
            ("{% bird:slot 'default' %}{% endbird:slot %}", {"slot": "test"}, "test"),
            ('{% bird:slot "default" %}{% endbird:slot %}', {"slot": "test"}, "test"),
            (
                "{% bird:slot name='default' %}{% endbird:slot %}",
                {"slot": "test"},
                "test",
            ),
            (
                '{% bird:slot name="default" %}{% endbird:slot %}',
                {"slot": "test"},
                "test",
            ),
            (
                "{% bird:slot name='not-default' %}{% endbird:slot %}",
                {"slot": "test"},
                "",
            ),
            (
                "{% bird:slot outer %}Outer {% bird:slot inner %}Inner{% endbird:slot %} Content{% endbird:slot %}",
                {"slots": {"outer": "Replaced Content"}},
                "Replaced Content",
            ),
            (
                "{% bird:slot outer %}Outer {% bird:slot inner %}Inner{% endbird:slot %} Content{% endbird:slot %}",
                {"slots": {"inner": "Replaced Content"}},
                "Outer Replaced Content Content",
            ),
            (
                "{% bird:slot outer %}Outer {% bird:slot inner %}Inner Default{% endbird:slot %} Content{% endbird:slot %}",
                {
                    "slots": {
                        "outer": "Replaced {% bird:slot inner %}{% endbird:slot %} Outer",
                        "inner": "Replaced Inner",
                    },
                },
                "Replaced Replaced Inner Outer",
            ),
            (
                "{{ slot }}",
                {
                    "slot": "Replaced {% bird:slot inner %}{% endbird:slot %} Outer",
                    "slots": {
                        "inner": "Replaced Inner",
                    },
                },
                "Replaced Replaced Inner Outer",
            ),
            (
                "{% bird:slot %}{% endbird:slot %}",
                {},
                "",
            ),
            (
                "{% bird:slot %}Default content{% endbird:slot %}",
                {},
                "Default content",
            ),
            (
                "{% bird:slot 'mal formed' %}{% endbird:slot %}",
                {"slots": {"mal formed": "content"}},
                "content",
            ),
            (
                "{% bird:slot CaseSensitive %}{% endbird:slot %}",
                {"slots": {"CaseSensitive": "Upper", "casesensitive": "Lower"}},
                "Upper",
            ),
            (
                "{% bird:slot %}{% endbird:slot %}",
                {"slots": {"default": 42}},
                "42",
            ),
            (
                "{% bird:slot %}{% endbird:slot %}",
                {"slots": {"default": "<b>Bold</b>"}},
                "<b>Bold</b>",
            ),
            (
                "{% bird:slot unicode_slot %}{% endbird:slot %}",
                {"slots": {"unicode_slot": "こんにちは"}},
                "こんにちは",
            ),
        ],
    )
    def test_rendering(self, template, context, expected, create_bird_template):
        create_bird_template("test", template)
        t = Template(f"{{% bird test %}}{template}{{% endbird %}}")
        rendered = t.render(context=Context(context))
        assert rendered == expected

    def test_too_many_args(self):
        with pytest.raises(TemplateSyntaxError):
            Template("{% bird:slot too many args %}{% endbird:slot %}")
